@@ -18,13 +18,13 @@ module Network.Wai.Middleware.Jsonp (jsonp) where
 import Network.Wai
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B8
-import Data.Enumerator (($$), enumList, Step (..), Enumerator, Iteratee, Enumeratee, joinI, checkDone, continue, Stream (..), (>>==))
-import Blaze.ByteString.Builder (copyByteString, Builder)
+import Blaze.ByteString.Builder (copyByteString)
 import Blaze.ByteString.Builder.Char8 (fromChar)
 import Data.Monoid (mappend)
 import Control.Monad (join)
 import Data.Maybe (fromMaybe)
 import qualified Data.ByteString as S
+import qualified Data.Conduit.List as CL
 
 -- | Wrap json responses in a jsonp callback.
 --
@@ -54,7 +54,6 @@ jsonp app env = do
         Nothing -> return res
         Just c -> go c res
   where
-    go c r@(ResponseFile _ hs _ _) = go' c r hs
     go c r@(ResponseBuilder s hs b) =
         case checkJSON hs of
             Nothing -> return r
@@ -63,39 +62,26 @@ jsonp app env = do
                 `mappend` fromChar '('
                 `mappend` b
                 `mappend` fromChar ')'
-    go c (ResponseEnumerator e) = addCallback c e
-    go' c r hs =
+    go c r =
         case checkJSON hs of
-            Just _ -> addCallback c $ responseEnumerator r
+            Just hs' -> addCallback c s hs' b
             Nothing -> return r
+      where
+        (s, hs, b) = responseSource r
+
     checkJSON hs =
         case lookup "Content-Type" hs of
             Just x
-                | B8.pack "application/json" `S.isPrefixOf` x -> Just $ fixHeaders hs
+                | B8.pack "application/json" `S.isPrefixOf` x ->
+                    Just $ fixHeaders hs
             _ -> Nothing
     fixHeaders = changeVal "Content-Type" "text/javascript"
-    addCallback :: B8.ByteString -> (forall a. ResponseEnumerator a)
-                -> Iteratee B8.ByteString IO Response
-    addCallback cb e =
-        return $ ResponseEnumerator $ helper
-      where
-        helper f =
-            e helper'
-          where
-            helper' s hs =
-                case checkJSON hs of
-                    Just hs' -> wrap $$ f s hs'
-                    Nothing -> f s hs
-        wrap :: Step Builder IO b -> Iteratee Builder IO b
-        wrap step = joinI $ after (enumList 1 [fromChar ')'])
-                 $$ enumList 1 [copyByteString cb, fromChar '('] step
-        after :: Enumerator Builder IO b -> Enumeratee Builder Builder IO b
-        after enum =
-            loop
-          where
-            loop = checkDone $ continue . step
-            step k EOF = enum (Continue k) >>== return
-            step k s = k s >>== loop
+
+    addCallback cb s hs b =
+        return $ ResponseSource s hs $
+            CL.sourceList [copyByteString cb `mappend` fromChar '(']
+            `mappend` b
+            `mappend` CL.sourceList [fromChar ')']
 
 changeVal :: Eq a
           => a
