@@ -5,6 +5,7 @@
 module Network.Wai.Parse
     ( parseHttpAccept
     , parseRequestBody
+    , conduitRequestBody
     , Sink (..)
     , lbsSink
     , tempFileSink
@@ -32,11 +33,11 @@ import Data.Function (on)
 import System.Directory (removeFile, getTemporaryDirectory)
 import System.IO (hClose, openBinaryTempFile, Handle)
 import Network.Wai
-import Data.Enumerator (Iteratee, yield)
-import qualified Data.Enumerator as E
-import qualified Data.Enumerator.List as EL
+import qualified Data.Conduit as C
+import qualified Data.Conduit.List as CL
 import Control.Monad.IO.Class (liftIO)
 import qualified Network.HTTP.Types as H
+import Data.Either (partitionEithers)
 
 uncons :: S.ByteString -> Maybe (Word8, S.ByteString)
 uncons s
@@ -129,16 +130,24 @@ type File y = (S.ByteString, FileInfo y)
 
 parseRequestBody :: Sink x y
                  -> Request
-                 -> Iteratee S.ByteString IO ([Param], [File y])
-parseRequestBody sink req = do
+                 -> C.Sink S.ByteString IO ([Param], [File y])
+parseRequestBody s r = fmap partitionEithers $ conduitRequestBody s r C.=$ CL.consume
+
+conduitRequestBody :: Sink x y
+                   -> Request
+                   -> C.Conduit S.ByteString IO (Either Param (File y))
+conduitRequestBody sink req = do
     case ctype of
-        Nothing -> return ([], [])
-        Just Nothing -> do -- url-encoded
+        Nothing -> C.Conduit $ return $ C.PreparedConduit
+            { C.conduitPush = \bs -> return $ C.Finished (Just bs) []
+            , C.conduitClose = return []
+            }
+        Just Nothing -> C.sequenceSink () $ \() -> do -- url-encoded
             -- NOTE: in general, url-encoded data will be in a single chunk.
             -- Therefore, I'm optimizing for the usual case by sticking with
             -- strict byte strings here.
-            bs <- EL.consume
-            return (H.parseSimpleQuery $ S.concat bs, [])
+            bs <- CL.consume
+            return $ C.Emit () $ map Left $ H.parseSimpleQuery $ S.concat bs
         Just (Just bound) -> -- multi-part
             let bound'' = S8.pack "--" `S.append` bound
              in parsePieces sink bound''
@@ -162,24 +171,20 @@ parseRequestBody sink req = do
                 Just x -> Just $ Just x
                 Nothing -> Nothing
 
-takeLine :: Iteratee S.ByteString IO (Maybe S.ByteString)
-takeLine = do
-    mbs <- EL.head
-    case mbs of
-        Nothing -> return Nothing
-        Just bs ->
-            let (x, y) = S.break (== 10) bs -- LF
-             in if S.null y
-                    then do
-                        x' <- takeLine
-                        case x' of
-                            Nothing -> return $ Just $ killCR x
-                            Just x'' -> return $ Just $ killCR $ S.append x x''
-                    else do
-                        E.yield () $ E.Chunks [S.drop 1 y]
-                        return $ Just $ killCR x
+takeLine :: C.Sink S.ByteString IO (Maybe S.ByteString)
+takeLine =
+    C.sinkState id push close
+  where
+    close _ = return Nothing
+    push front bs = do
+        let (x, y) = S.break (== 10) $ front bs -- LF
+         in if S.null y
+                then return (S.append x, C.Processing)
+                else do
+                    let lo = if S.length y > 1 then Just (S.drop 1 y) else Nothing
+                    return (error "takeLine", C.Done lo $ Just $ killCR x)
 
-takeLines :: Iteratee S.ByteString IO [S.ByteString]
+takeLines :: C.Sink S.ByteString IO [S.ByteString]
 takeLines = do
     res <- takeLine
     case res of
@@ -191,12 +196,18 @@ takeLines = do
                 return $ l : ls
 
 parsePieces :: Sink x y -> S.ByteString
-            -> Iteratee S.ByteString IO ([Param], [File y])
-parsePieces sink bound = do
+            -> C.Conduit S.ByteString IO (Either Param (File y))
+parsePieces sink bound = C.sequenceSink True (parsePiecesSink sink bound)
+
+parsePiecesSink :: Sink x y
+                -> S.ByteString
+                -> C.SequencedSink Bool S.ByteString IO (Either Param (File y))
+parsePiecesSink _ _ False = return C.Stop
+parsePiecesSink sink bound True = do
     _boundLine <- takeLine
     res' <- takeLines
     case res' of
-        [] -> return ([], [])
+        [] -> return C.Stop
         _ -> do
             let ls' = map parsePair res'
             let x = do
@@ -214,11 +225,7 @@ parsePieces sink bound = do
                     y <- liftIO $ sinkClose sink seed'
                     let fi = FileInfo filename ct y
                     let y' = (name, fi)
-                    (xs, ys) <-
-                        if wasFound
-                            then parsePieces sink bound
-                            else return ([], [])
-                    return (xs, y' : ys)
+                    return $ C.Emit wasFound [Right y']
                 Just (_ct, name, Nothing) -> do
                     let seed = id
                     let iter front bs = return $ front . (:) bs
@@ -226,19 +233,13 @@ parsePieces sink bound = do
                         sinkTillBound bound iter seed
                     let bs = S.concat $ front []
                     let x' = (name, qsDecode bs)
-                    (xs, ys) <-
-                        if wasFound
-                            then parsePieces sink bound
-                            else return ([], [])
-                    return (x' : xs, ys)
+                    return $ C.Emit wasFound [Left x']
                 _ -> do
                     -- ignore this part
                     let seed = ()
                         iter () _ = return ()
                     ((), wasFound) <- sinkTillBound bound iter seed
-                    if wasFound
-                        then parsePieces sink bound
-                        else return ([], [])
+                    return $ C.Emit wasFound []
   where
     contDisp = S8.pack "Content-Disposition"
     contType = S8.pack "Content-Type"
@@ -271,56 +272,32 @@ findBound b bs = go [0..S.length bs - 1]
 sinkTillBound :: S.ByteString
               -> (x -> S.ByteString -> IO x)
               -> x
-              -> Iteratee S.ByteString IO (x, Bool)
-sinkTillBound bound iter seed = do
-    mbs <- EL.head
-    case mbs of
-        Nothing -> return (seed, False)
-        Just bs -> go bs
+              -> C.Sink S.ByteString IO (x, Bool)
+sinkTillBound bound iter seed0 = C.sinkState
+    (id, seed0)
+    push
+    close
   where
-    go bs =
+    close (front, seed) = do
+        seed' <- liftIO $ iter seed $ front S.empty
+        return (seed', False)
+    push (front, seed) bs' = do
+        let bs = front bs'
         case findBound bound bs of
             FoundBound before after -> do
                 let before' = killCRLF before
                 seed' <- liftIO $ iter seed before'
-                yield () $ E.Chunks [after]
-                return (seed', True)
-            PartialBound -> do
-                mbs <- EL.head
-                case mbs of
-                    Nothing -> do
-                        seed' <- liftIO $ iter seed bs
-                        return (seed', False)
-                    Just bs2 -> do
-                        let bs' = bs `S.append` bs2
-                        yield () $ E.Chunks [bs']
-                        sinkTillBound bound iter seed
+                return (undefined, C.Done (Just after) (seed', True))
             NoBound -> do
-                mbs <- EL.head
-                case mbs of
-                    Nothing -> do
-                        seed' <- liftIO $ iter seed bs
-                        sinkTillBound bound iter seed'
-                    Just bs' -> do
-                        -- this funny bit is to catch when there's a
-                        -- newline at the end of the previous chunk
-                        (seed', bs'') <-
-                            if not (S8.null bs) && S8.last bs `elem` "\n\r"
-                                then do
-                                    let (front, back) =
-                                            S.splitAt (S.length bs - 2) bs
-                                    seed' <- liftIO $ iter seed front
-                                    return (seed', back `S.append` bs')
-                                else do
-                                    seed' <- liftIO $ iter seed bs
-                                    return (seed', bs')
-                        yield () $ E.Chunks [bs'']
-                        sinkTillBound bound iter seed'
-        {-
-        NoBound -> do
-            case msrc of
-                        Just (bs', src') -> do
-        -}
+                -- don't emit newlines, in case it's part of a bound
+                let (toEmit, front') =
+                        if not (S8.null bs) && S8.last bs `elem` "\r\n"
+                            then let (x, y) = S.splitAt (S.length bs - 2) bs
+                                  in (x, S.append y)
+                            else (bs, id)
+                seed' <- liftIO $ iter seed toEmit
+                return ((front', seed'), C.Processing)
+            PartialBound -> return ((S.append bs, seed), C.Processing)
 
 parseAttrs :: S.ByteString -> [(S.ByteString, S.ByteString)]
 parseAttrs = map go . S.split 59 -- semicolon
